/************************************************************************/
/*	
*	WZORZEC PROJEKTOWY KOMENDY
*	
*	Piotr Falgowski G1SST
*
*	Opis: 
*
*	Wzorzec KOMENDY jest wzorcem projektowym nalecym do wzorcw czynnociowym. 
*	Pozwala enkapsulowa czynnoci i ich parametry w postaci obiektw. Dziki temu 
*	obiekt moe by gromadzony i wpuszczany w obieg tak jak inne obiekty. Kluczem do 
*	tego wzoru jest abstrakcyjna klasa Command, w ktrej deklarujemy interfejs do 
*	wykonywania operacji. W najprostszej formie ten interfejs zawiera operacj Execute.
*	Konkretne podklasy Command precyzuj akcje odbiorcy, tworzc magazyn zada 
*	i przygotowuj narzdzia do wykonywania zada.
*
*             ____________________               ______________________                
*             | KlientWywolujcy |-------------->|   <<Interfejs>>    |                
*             |__________________|               |       Command      |                
*             |__________________|               |____________________|                 
*             |__________________|               |   + execute()=0    |                 
*                                                |____________________|                 
*                                                    /|\        /|\						
*                                       ______________|          |					
*   ______________________       _______|______________      _________|________			
*   |  Klient Odbiorca   |       |  Konkretna Komenda |      |  Inna Komenda  |		
*   |____________________|<------|____________________|      |________________|		
*   |____________________|       |____________________|      |________________|		
*   |____________________|       |  + execute()       |      | + execute()    |			
*                                |____________________|      |________________|		
*
*                                                                                  
*	Wzorzec komendy naley uywa wszdzie tam gdy:
*		-	parametryzujesz obiekty poprzez akcje ktre wykonuj,
*		-	precyzujesz, ustawiasz w kolejce i wykonujesz dania w rnym czasie,
*		-	chcesz mie moliwo cofa, przywraca akcje,
*		-	chcesz mie moliwo rejestrowania zmian tak by mogy by one ponownie
*			wykorzystane w przypadku gdy system przestanie dziaa,
*
*/
/************************************************************************/


// Przyklad uycia wzorca komendy do stworzenia "elektornicznego organizera";) - przyklad cofania i przywracania zmian

#include <iostream>
#include <vector>
#include <string>

using namespace std;


//Klasa - interfejs po ktrej bd dziedziczy konkretne komendy
class Command{
public:

	// Metoda execute - wykonanie komendy.
	virtual void execute() = 0;
	// Metoda undo - potrzebna gdy chcemy mie funkcje cofania zmian.
	virtual void undo() = 0;
	virtual ~Command(){};
};

//Klasa addMeeting - konkretyzacja klasy Command - dodaje spotkanie do organizera. 
class addMeeting : public Command {
public:

	// konstruktor przyjmuje argumenty z jakimi ma by pniej wykonana komenda,
	// oraz adres obiektu na rzecz ktrego bdzie wykonana
	addMeeting(string person, string time, string* doc){
		_person = person;
		_time = time;
		_doc = doc;
	}

	// Metoda execute - dodaje do "organizera" spotaknie, oraz robi jego kopie przed zmian.
	void execute(){
		_prevDoc = *_doc;
		*_doc += ("Spotkanie z " +_person +" o godzinie: "+ _time + "\n");
	}

	// Metoda undo przywraca "organizer" do stanu sprzed wykonania komendy
	void undo(){
		*_doc = _prevDoc;
	}
private:
	string _person;		// parametr komendy
	string _time;		// parametr komendy
	string _prevDoc;	// kopia obiektu przed wykonaniem komendy
	string* _doc;		// wskanik na obiekt na rzecz ktrego bdzie wykonywana komenda
};

// Klasa addActivity -	podobna jak addMeeting rwnie dziedziczy po Command, 
//						wykonuje tylko inn operacje na obiekcie. Struktura pl 
//						i metod jak wyej.
class addActivity : public Command {
public:
	addActivity(string activity, string time, string* doc){
		_activity = activity;
		_time = time;
		_doc = doc;
	}
	void execute(){
		_prevDoc = *_doc;
		*_doc += (_activity +" o godzinie: "+ _time + "\n");
	}
	void undo(){
		*_doc = _prevDoc;
	}
private:
	string _activity;
	string _time;
	string _prevDoc;
	string* _doc;
};


// Klasa CmdStack reprezentuje wektor komend.
class CmdStack{
public:

	// Metoda add dodaj komend oraz j wykonuje.
	void add(Command *c) {
		c->execute();
		commands.push_back(c);
		commands_redo.clear();
	}

	// Metoda undo potrafi cofn dokonane zmiany - zmiana w ty.
	void undo(){
		if(commands.size() > 1) {
			commands_redo.push_back(commands[commands.size()-1]);
			commands[commands.size()-1]->undo();
			commands.pop_back();
		}
		else {
			std::cout << "Nie mozna wykonac undo" << std::endl;
		}
	}

	// Metoda redo przywraca stan przed wykonaniem undo - zmiana w przd.
	void redo(){
		if(commands_redo.size() >= 1) {
			commands.push_back(commands_redo[commands_redo.size()-1]);
			commands_redo[commands_redo.size()-1]->execute();
			commands_redo.pop_back();
		}	
		else {
			std::cout << "Nie mozna wykonac redo" << std::endl;
		}
	}
private:
	vector<Command*> commands; //wektor trzymajcy komendy - zmiany w ty
	vector<Command*> commands_redo; // wektor trzymajcy komendy - zmiany w przd
};
int main() {
	CmdStack list;
	string plan; // nasz "organizer" ;)

	std::cout<<"Dodajemy spotkanie z Johnem."<<std::endl;
	list.add(new addMeeting("John Shith", "8:00 - 8:45", &plan));

	std::cout<<"Dodajemy spotkanie z Arnoldem."<<std::endl;
	list.add(new addMeeting("Arnold S.", "9:00 - 15:00", &plan));

	std::cout <<"Nasz plan dnia:"<<std::endl;
	std::cout << plan<<endl;

	std::cout <<"Cofamy ostania zmiane w organizerze"<<std::endl;
	list.undo();
	std::cout << "Nasz plan po cofnieciu ostatniej zmiany:"<<endl;
	std::cout << plan<<endl;

	std::cout <<"Przywracamy ostania zmiane w organizerze"<<std::endl;
	list.redo();
	std::cout << "Nasz plan po przywroceniu ostatniej zmiany:"<<endl;
	std::cout << plan<<endl;

	std::cout<<"Dodajemy koszykowke."<<std::endl;
	list.add(new addActivity("Koszykowka", "16:00 - 18:00", &plan));
	std::cout <<"Nasz ostateczny plan dnia:"<<std::endl;
	std::cout << plan<<endl;

	return 0;
}

